# Copyright 2022 (c) Anna Schumaker.
"""A custom Gtk.HeaderBar configured for our application."""
import pathlib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Adw
from .. import db
from .. import buttons
from . import open
from . import replaygain
from . import volume
if __debug__:
    from . import settings

SUBTITLE = "The Cheesy Music Player"


def _volume_icon(vol: float) -> str:
    if vol == 0.0:
        return "audio-volume-muted-symbolic"
    if vol <= 1/3:
        return "audio-volume-low-symbolic"
    if vol <= 2/3:
        return "audio-volume-medium-symbolic"
    return "audio-volume-high-symbolic"


class Header(Gtk.HeaderBar):
    """Our custom Gtk.HeaderBar containing window title and volume controls."""

    sql = GObject.Property(type=db.Connection)
    title = GObject.Property(type=str)
    subtitle = GObject.Property(type=str)
    rg_enabled = GObject.Property(type=bool, default=False)
    rg_mode = GObject.Property(type=str, default="auto")
    volume = GObject.Property(type=float, default=1.0)

    def __init__(self, sql: db.Connection, title: str):
        """Initialize the HeaderBar."""
        super().__init__(title=title, subtitle=SUBTITLE, sql=sql)
        self._open = open.Button()
        self._title = Adw.WindowTitle(title=self.title, subtitle=self.subtitle)
        self._volume = volume.Controls()
        self._replaygain = replaygain.Selector()

        self._box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        self._box.append(self._volume)
        self._box.append(Gtk.Separator())
        self._box.append(self._replaygain)

        icon = _volume_icon(self.volume)
        self._button = buttons.PopoverButton(popover_child=self._box,
                                             icon_name=icon)

        self.bind_property("title", self._title, "title")
        self.bind_property("subtitle", self._title, "subtitle")
        self.bind_property("rg-enabled", self._replaygain, "enabled",
                           GObject.BindingFlags.BIDIRECTIONAL)
        self.bind_property("rg-mode", self._replaygain, "mode",
                           GObject.BindingFlags.BIDIRECTIONAL)
        self.bind_property("volume", self._volume, "volume",
                           GObject.BindingFlags.BIDIRECTIONAL)

        self.pack_start(self._open)
        if __debug__:
            self._window = settings.Window(sql)
            self._settings = Gtk.Button.new_from_icon_name("settings-symbolic")
            self._settings.connect("clicked", self.__run_settings)
            self.pack_start(self._settings)

        self.pack_end(self._button)
        self.set_title_widget(self._title)

        self._open.connect("track-requested", self.__track_requested)
        self.connect("notify::volume", self.__notify_volume)

    def __run_settings(self, button: Gtk.Button) -> None:
        if __debug__:
            self._window.present()

    def __notify_volume(self, header, param) -> None:
        self._button.set_icon_name(_volume_icon(self.volume))

    def __track_requested(self, button: open.Button,
                          path: pathlib.Path) -> None:
        self.emit("track-requested", path)

    @GObject.Signal(arg_types=(GObject.TYPE_PYOBJECT,))
    def track_requested(self, path: pathlib.Path) -> None:
        """Signal that a track has been requested."""
