# Copyright 2022 (c) Anna Schumaker.
"""A card for displaying the list of playlists."""
from gi.repository import GObject
from gi.repository import Gtk
from . import artist
from . import decade
from . import genre
from . import library
from . import playlist
from . import section
from .. import db
from .. import entry


class Card(Gtk.Box):
    """Our playlist Sidebar."""

    sql = GObject.Property(type=db.Connection)
    show_all_artists = GObject.Property(type=bool, default=False)
    selected_playlist = GObject.Property(type=db.playlist.Playlist)

    def __init__(self, sql: db.Connection, **kwargs):
        """Set up the Sidebar widget."""
        super().__init__(sql=sql, orientation=Gtk.Orientation.VERTICAL,
                         sensitive=False, **kwargs)
        self._filter = entry.Filter("playlists")
        self._playlists = playlist.Section(self.sql.playlists)
        self._artists = artist.Section(self.sql.artists, self.sql.albums)
        self._genres = genre.Section(self.sql.genres)
        self._decades = decade.Section(self.sql.decades, self.sql.years)
        self._libraries = library.Section(self.sql.libraries)
        self._group = section.Group(sql)

        self.append(self._filter)

        for sect in [self._playlists, self._artists, self._genres,
                     self._decades, self._libraries]:
            self.append(sect)
            self._group.add(sect)

        self._group.bind_property("selected-playlist",
                                  self, "selected-playlist")
        self.bind_property("show-all-artists", self._artists, "show-all",
                           GObject.BindingFlags.BIDIRECTIONAL)

        self._filter.connect("search-changed", self.__search_changed)
        self.sql.connect("table-loaded", self.__table_loaded)

        self.add_css_class("background")
        self.add_css_class("linked")
        self.add_css_class("card")

    def __search_changed(self, entry: entry.Filter) -> None:
        self.sql.filter(entry.get_query())

    def __table_loaded(self, sql: db.Connection, table: db.table.Table):
        if self.get_sensitive() is False:
            if False not in {tbl.loaded for tbl in sql.playlist_tables()}:
                self.set_sensitive(True)
                self.select_playlist(sql.active_playlist)
                if len(sql.libraries) == 0:
                    self._libraries.extra_widget.emit("clicked")

    def select_playlist(self, playlist: db.playlist.Playlist) -> None:
        """Set the current active playlist."""
        if playlist is not None:
            match playlist.table:
                case self.sql.playlists:
                    section = self._playlists
                case self.sql.artists | self.sql.albums | self.sql.media:
                    section = self._artists
                case self.sql.genres:
                    section = self._genres
                case self.sql.decades | self.sql.years:
                    section = self._decades
                case self.sql.libraries:
                    section = self._libraries

            section.active = True
            section.select_playlist(playlist)
