# Copyright 2022 (c) Anna Schumaker.
"""Tests our application header."""
import emmental
import tests.util
import pathlib
import unittest.mock
from gi.repository import Gtk
from gi.repository import Adw


class TestHeader(tests.util.TestCase):
    """Test case for our custom Gtk.HeaderBar."""

    def setUp(self):
        """Set up common variables."""
        super().setUp()
        self.header = emmental.header.Header(sql=self.sql, title="Test Title")

    def test_constants(self):
        """Check that constants are set correctly."""
        self.assertEqual(emmental.header.SUBTITLE, "The Cheesy Music Player")

    def test_title(self):
        """Check that the title is set correctly."""
        self.assertIsInstance(self.header, Gtk.HeaderBar)
        self.assertIsInstance(self.header._title, Adw.WindowTitle)

        self.assertEqual(self.header.get_title_widget(), self.header._title)
        self.assertEqual(self.header.title, "Test Title")
        self.assertEqual(self.header._title.get_title(), "Test Title")

        self.assertEqual(self.header.subtitle, emmental.header.SUBTITLE)
        self.assertEqual(self.header._title.get_subtitle(),
                         emmental.header.SUBTITLE)

    def test_open(self):
        """Check that the Open button works as expected."""
        self.assertIsInstance(self.header._open, emmental.header.open.Button)

        signal = unittest.mock.Mock()
        self.header.connect("track-requested", signal)
        self.header._open.emit("track-requested", pathlib.Path("/a/b/c/1.ogg"))
        signal.assert_called_with(self.header, pathlib.Path("/a/b/c/1.ogg"))

    def test_settings(self):
        """Check that the Settings window is set up correctly."""
        self.assertIsInstance(self.header._settings, Gtk.Button)
        self.assertIsInstance(self.header._window,
                              emmental.header.settings.Window)

        self.assertEqual(self.header.sql, self.sql)
        self.assertEqual(self.header._settings.get_icon_name(),
                         "settings-symbolic")

        with unittest.mock.patch.object(self.header._window,
                                        "present") as mock_present:
            self.header._settings.emit("clicked")
            mock_present.assert_called()

    def test_volume(self):
        """Check that volume widgets work as expected."""
        self.assertIsInstance(self.header._volume,
                              emmental.header.volume.Controls)
        self.assertEqual(self.header.volume, 1.0)

        for i, vol in [(x, x/10) for x in range(11)]:
            with self.subTest(volume=vol):
                widget = self.header if i % 2 else self.header._volume
                match i:
                    case 0: icon = "muted"
                    case 1 | 2 | 3: icon = "low"
                    case 4 | 5 | 6: icon = "medium"
                    case _: icon = "high"

                widget.volume = vol
                self.assertEqual(self.header.volume, vol)
                self.assertEqual(self.header._volume.volume, vol)
                self.assertEqual(self.header._button.get_icon_name(),
                                 f"audio-volume-{icon}-symbolic")

    def test_replaygain(self):
        """Test that we can configure ReplayGain as expected."""
        self.assertIsInstance(self.header._replaygain,
                              emmental.header.replaygain.Selector)
        self.assertFalse(self.header.rg_enabled)
        self.assertEqual(self.header.rg_mode, "auto")

        self.header.rg_enabled = True
        self.header.rg_mode = "track"
        self.assertTrue(self.header._replaygain.enabled)
        self.assertEqual(self.header._replaygain.mode, "track")

        self.header._replaygain.enabled = False
        self.header._replaygain.mode = "album"
        self.assertFalse(self.header.rg_enabled)
        self.assertEqual(self.header.rg_mode, "album")

    def test_popover(self):
        """Check that the menu popover was set up correctly."""
        self.assertIsInstance(self.header._button,
                              emmental.buttons.PopoverButton)
        self.assertIsInstance(self.header._box, Gtk.Box)

        self.assertEqual(self.header._box.get_orientation(),
                         Gtk.Orientation.VERTICAL)
        self.assertEqual(self.header._box.get_spacing(), 0)

        self.assertEqual(self.header._button.get_icon_name(),
                         "audio-volume-high-symbolic")
        self.assertEqual(self.header._button.popover_child, self.header._box)
        self.assertEqual(self.header._box.get_first_child(),
                         self.header._volume)

        sep = self.header._volume.get_next_sibling()
        self.assertIsInstance(sep, Gtk.Separator)
        self.assertEqual(sep.get_orientation(), Gtk.Orientation.HORIZONTAL)
        self.assertEqual(sep.get_next_sibling(), self.header._replaygain)
