# Copyright 2022 (c) Anna Schumaker.
"""Tests our title widgets."""
import unittest
import emmental.sidebar.title
from gi.repository import Pango
from gi.repository import Gtk


class TestTitle(unittest.TestCase):
    """Tests the row title widget."""

    def setUp(self):
        """Initialize common variables."""
        self.title = emmental.sidebar.title.Title()

    def test_init(self):
        """Test that the row Title is set up properly."""
        self.assertIsInstance(self.title, Gtk.Box)
        self.assertEqual(self.title.get_orientation(),
                         Gtk.Orientation.VERTICAL)
        self.assertEqual(self.title.get_valign(), Gtk.Align.CENTER)
        self.assertEqual(self.title.get_margin_top(), 6)
        self.assertEqual(self.title.get_margin_bottom(), 6)
        self.assertTrue(self.title.get_hexpand())

    def test_title(self):
        """Test the title property."""
        self.assertIsInstance(self.title._title, Gtk.Label)
        self.assertEqual(self.title.title, "")
        self.assertEqual(self.title.get_first_child(), self.title._title)
        self.assertIsNone(self.title.get_tooltip_text())

        self.assertEqual(self.title._title.get_text(), "")
        self.assertEqual(self.title._title.get_xalign(), 0)
        self.assertEqual(self.title._title.get_ellipsize(),
                         Pango.EllipsizeMode.END)
        self.assertTrue(self.title._title.has_css_class("header"))

        self.title.title = "Test Title"
        self.assertEqual(self.title._title.get_text(), "Test Title")
        self.assertEqual(self.title.get_tooltip_text(), "Test Title")

        title2 = emmental.sidebar.title.Title(title="Other Title")
        self.assertEqual(title2._title.get_text(), "Other Title")
        self.assertEqual(title2.get_tooltip_text(), "Other Title")

    def test_subtitle(self):
        """Test the subtitle property."""
        self.assertIsInstance(self.title._subtitle, Gtk.Label)
        self.assertEqual(self.title.subtitle, "")
        self.assertEqual(self.title._title.get_next_sibling(),
                         self.title._subtitle)

        self.assertEqual(self.title._subtitle.get_text(), "")
        self.assertEqual(self.title._subtitle.get_xalign(), 0)
        self.assertEqual(self.title._subtitle.get_ellipsize(),
                         Pango.EllipsizeMode.END)
        self.assertTrue(self.title._subtitle.has_css_class("caption"))

        self.title.subtitle = "Test Subtitle"
        self.assertEqual(self.title._subtitle.get_text(), "Test Subtitle")

        title2 = emmental.sidebar.title.Title(subtitle="Other Subtitle")
        self.assertEqual(title2._subtitle.get_text(), "Other Subtitle")


class TestPlaylistTitle(unittest.TestCase):
    """Test our title widget configured for displaying playlist size."""

    def setUp(self):
        """Set up common variables."""
        self.title = emmental.sidebar.title.PlaylistTitle()

    def test_init(self):
        """Test that the PlaylistTitle is configured correctly."""
        self.assertIsInstance(self.title, emmental.sidebar.title.Title)
        self.assertEqual(self.title.subtitle, "-- tracks")
        self.assertEqual(self.title.count, 0)

    def test_count(self):
        """Test changing the count property."""
        self.title.count = 0
        self.assertEqual(self.title.subtitle, "0 tracks")

        self.title.count = 1
        self.assertEqual(self.title.subtitle, "1 track")

        self.title.count = 2
        self.assertEqual(self.title.subtitle, "2 tracks")


class TestEditableTitle(unittest.TestCase):
    """Tests the editable title widget."""

    def setUp(self):
        """Set up common variables."""
        self.title = emmental.sidebar.title.EditableTitle()

    def test_init(self):
        """Test that the editable title widget is set up properly."""
        self.assertIsInstance(self.title, Gtk.Stack)
        self.assertIsNotNone(self.title.get_page(self.title._title))
        self.assertIsNotNone(self.title.get_page(self.title._entry))

    def test_playlist_title(self):
        """Test the playlist title child."""
        self.assertIsInstance(self.title._title,
                              emmental.sidebar.title.PlaylistTitle)
        self.assertEqual(self.title.title, "")
        self.assertEqual(self.title.count, 0)

        self.title.title = "Test Title"
        self.assertEqual(self.title._title.title, "Test Title")
        self.assertEqual(self.title._entry.get_text(), "Test Title")

        self.title.count = 42
        self.assertEqual(self.title._title.count, 42)

        title2 = emmental.sidebar.title.EditableTitle(title="Test Title")
        self.assertEqual(title2._title.title, "Test Title")
        self.assertEqual(title2._entry.get_text(), "Test Title")

    def test_visible_child(self):
        """Test switching to the Entry for editing."""
        self.assertEqual(self.title.get_visible_child(), self.title._title)
        self.assertFalse(self.title.editing)

        self.title.editing = True
        self.assertFalse(self.title.editing)

        self.title.editable = True
        self.title.editing = True
        self.assertTrue(self.title.editing)
        self.assertEqual(self.title.get_visible_child(), self.title._entry)

        self.title.editing = False
        self.assertEqual(self.title.get_visible_child(), self.title._title)

    def test_cancel(self):
        """Test the cancel Shortcut for editing."""
        self.assertIsInstance(self.title._shortcut, Gtk.ShortcutController)
        self.assertIsInstance(self.title._cancel, Gtk.Shortcut)
        self.assertIsInstance(self.title._trigger, Gtk.KeyvalTrigger)
        self.assertIsInstance(self.title._action, Gtk.CallbackAction)

        self.assertEqual(self.title._cancel.get_trigger(), self.title._trigger)
        self.assertEqual(self.title._cancel.get_action(), self.title._action)

        self.assertEqual(self.title._trigger.get_keyval(),
                         Gtk.accelerator_parse("Escape").accelerator_key)
        self.assertEqual(self.title._trigger.get_modifiers(), 0)

        self.assertIn(self.title._cancel, self.title._shortcut)
        self.assertIn(self.title._shortcut, self.title.observe_controllers())

        self.title.editing = True
        self.title._action.activate(0, self.title, None)
        self.assertFalse(self.title.editing)

    def test_editing(self):
        """Test editing the title label."""
        self.assertIsInstance(self.title._entry, Gtk.Text)
        self.assertIsInstance(self.title._long_press, Gtk.GestureLongPress)
        self.assertIn(self.title._long_press, self.title.observe_controllers())

        self.title._long_press.emit("pressed", 0, 0)
        self.assertFalse(self.title.editing)

        self.title.editable = True
        self.title._long_press.emit("pressed", 0, 0)
        self.assertTrue(self.title.editing)

        callback = unittest.mock.Mock()
        self.title.connect("request-rename", callback)
        self.title._entry.set_text("New Name")
        self.title._entry.emit("activate")
        self.assertFalse(self.title.editing)
        callback.assert_called_with(self.title, "New Name")
