# Copyright 2022 (c) Anna Schumaker.
"""Test saving and loading Emmental settings."""
import io
import unittest
import unittest.mock
import emmental


@unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
class TestSettings(unittest.TestCase):
    """Emmental settings test case."""

    @unittest.mock.patch("gi.repository.Adw.Application.add_window")
    @unittest.mock.patch("gi.repository.Adw.Application.do_startup")
    def setUp(self, mock_startup: unittest.mock.Mock,
              mock_add_window: unittest.mock.Mock):
        """Set up common variables."""
        self.app = emmental.Application()
        with unittest.mock.patch("sys.stdout"):
            self.app.do_startup()
        self.settings = self.app.db.settings
        self.win = self.app.win
        self.player = self.app.player

    def tearDown(self):
        """Clean up."""
        self.app.do_shutdown()

    def test_save_window_size(self, new_callable=io.StringIO):
        """Check saving and loading window size from the database."""
        self.assertEqual(self.settings["window.width"], 1600)
        self.assertEqual(self.settings["window.height"], 900)

        self.win.set_default_size(100, 200)
        self.assertEqual(self.settings["window.width"], 100)
        self.assertEqual(self.settings["window.height"], 200)

        win = self.app.build_window()
        self.assertEqual(win.get_default_size(), (100, 200))

    def test_save_volume(self, mock_stdout: io.StringIO):
        """Check saving and loading volume from the database."""
        self.assertEqual(self.settings["audio.volume"], 1.0)
        self.assertEqual(self.player.volume, 1.0)

        self.win.header.volume = 0.5
        self.assertEqual(self.settings["audio.volume"], 0.5)
        self.assertEqual(self.player.volume, 0.5)

        self.player.volume = 0.0
        self.assertEqual(self.app.build_header().volume, 0.5)
        self.assertEqual(self.player.volume, 0.5)

    def test_save_replaygain(self, mock_stdout: io.StringIO):
        """Check saving and loading replaygain state from the database."""
        self.assertFalse(self.settings["audio.replaygain.enabled"])
        self.assertEqual(self.settings["audio.replaygain.mode"], "auto")
        self.assertEqual(self.player.get_replaygain(), (False, None))

        self.win.header.rg_enabled = True
        self.assertTrue(self.settings["audio.replaygain.enabled"])

        self.win.header.rg_mode = "track"
        self.assertEqual(self.settings["audio.replaygain.mode"], "track")
        self.assertEqual(self.player.get_replaygain(), (True, "track"))

        self.win.header.rg_mode = "album"
        self.assertEqual(self.settings["audio.replaygain.mode"], "album")
        self.assertEqual(self.player.get_replaygain(), (True, "album"))

        self.player.set_replaygain(False, None)
        header = self.app.build_header()
        self.assertTrue(header.rg_enabled)
        self.assertEqual(header.rg_mode, "album")
        self.assertEqual(self.player.get_replaygain(), (True, "album"))

        header.rg_enabled = False
        self.assertFalse(self.settings["audio.replaygain.enabled"])
        self.assertEqual(self.settings["audio.replaygain.mode"], "album")
        self.assertEqual(self.player.get_replaygain(), (False, None))

    def test_save_nowplaying_size(self, mock_stdout: io.StringIO):
        """Check saving and loading the nowplaying widget size."""
        self.assertEqual(self.win.now_playing_size, 250)
        self.assertEqual(self.settings["now-playing.size"], 250)

        self.win.now_playing_size = 400
        self.assertEqual(self.settings["now-playing.size"], 400)

        self.assertEqual(self.app.build_window().now_playing_size, 400)

    def test_save_nowplaying_prefer_artist(self, mock_stdout: io.StringIO):
        """Check saving and loading the prefer-artist setting."""
        self.assertTrue(self.win.now_playing.prefer_artist)
        self.assertTrue(self.settings["now-playing.prefer-artist"])

        self.win.now_playing.prefer_artist = False
        self.assertFalse(self.settings["now-playing.prefer-artist"])

        self.assertFalse(self.app.build_window().now_playing.prefer_artist)

    def test_save_sidebar_size(self, mock_stdout: io.StringIO):
        """Check saving and loading the sidebar widget size."""
        self.assertEqual(self.win.sidebar_size, 300)
        self.assertEqual(self.settings["sidebar.size"], 300)

        self.win.sidebar_size = 400
        self.assertEqual(self.settings["sidebar.size"], 400)

        self.assertEqual(self.app.build_window().sidebar_size, 400)

    def test_save_sidebar_show_all_artists(self, mock_stdout: io.StringIO):
        """Check saving and loading the show-all artists setting."""
        self.assertFalse(self.win.sidebar.show_all_artists)
        self.assertFalse(self.settings["sidebar.artists.show-all"])

        self.win.sidebar.show_all_artists = True
        self.assertTrue(self.settings["sidebar.artists.show-all"])

        self.assertTrue(self.app.build_window().sidebar.show_all_artists)

    def test_save_tracklist_column_width(self, mock_stdout: io.StringIO):
        """Test saving tracklist column widths."""
        self.assertEqual(self.settings["tracklist.art.size"], -1)
        self.assertEqual(self.settings["tracklist.fav.size"], -1)
        self.assertEqual(self.settings["tracklist.track.size"], 55)
        self.assertEqual(self.settings["tracklist.title.size"], 300)
        self.assertEqual(self.settings["tracklist.length.size"], -1)
        self.assertEqual(self.settings["tracklist.artist.size"], 250)
        self.assertEqual(self.settings["tracklist.album.size"], 350)
        self.assertEqual(self.settings["tracklist.album-artist.size"], 250)
        self.assertEqual(self.settings["tracklist.release.size"], 115)
        self.assertEqual(self.settings["tracklist.play-count.size"], 135)
        self.assertEqual(self.settings["tracklist.last-started.size"], 250)
        self.assertEqual(self.settings["tracklist.last-played.size"], 250)
        self.assertEqual(self.settings["tracklist.filepath.size"], -1)

        for column in self.win.tracklist.columns:
            column.set_fixed_width(123)

        self.assertEqual(self.settings["tracklist.art.size"], 123)
        self.assertEqual(self.settings["tracklist.fav.size"], 123)
        self.assertEqual(self.settings["tracklist.track.size"], 123)
        self.assertEqual(self.settings["tracklist.title.size"], 123)
        self.assertEqual(self.settings["tracklist.length.size"], 123)
        self.assertEqual(self.settings["tracklist.artist.size"], 123)
        self.assertEqual(self.settings["tracklist.album.size"], 123)
        self.assertEqual(self.settings["tracklist.album-artist.size"], 123)
        self.assertEqual(self.settings["tracklist.release.size"], 123)
        self.assertEqual(self.settings["tracklist.play-count.size"], 123)
        self.assertEqual(self.settings["tracklist.last-started.size"], 123)
        self.assertEqual(self.settings["tracklist.last-played.size"], 123)
        self.assertEqual(self.settings["tracklist.filepath.size"], 123)

    def test_save_tracklist_column_visible(self, mock_stdout: io.StringIO):
        """Test saving tracklist column visibility."""
        self.assertTrue(self.settings["tracklist.art.visible"])
        self.assertTrue(self.settings["tracklist.fav.visible"])
        self.assertTrue(self.settings["tracklist.track.visible"])
        self.assertTrue(self.settings["tracklist.title.visible"])
        self.assertTrue(self.settings["tracklist.length.visible"])
        self.assertTrue(self.settings["tracklist.artist.visible"])
        self.assertTrue(self.settings["tracklist.album.visible"])
        self.assertFalse(self.settings["tracklist.album-artist.visible"])
        self.assertTrue(self.settings["tracklist.release.visible"])
        self.assertTrue(self.settings["tracklist.play-count.visible"])
        self.assertFalse(self.settings["tracklist.last-started.visible"])
        self.assertTrue(self.settings["tracklist.last-played.visible"])
        self.assertFalse(self.settings["tracklist.filepath.visible"])

        for column in self.win.tracklist.columns:
            column.set_visible(not column.get_visible())

        self.assertFalse(self.settings["tracklist.art.visible"])
        self.assertFalse(self.settings["tracklist.fav.visible"])
        self.assertFalse(self.settings["tracklist.track.visible"])
        self.assertFalse(self.settings["tracklist.title.visible"])
        self.assertFalse(self.settings["tracklist.length.visible"])
        self.assertFalse(self.settings["tracklist.artist.visible"])
        self.assertFalse(self.settings["tracklist.album.visible"])
        self.assertTrue(self.settings["tracklist.album-artist.visible"])
        self.assertFalse(self.settings["tracklist.release.visible"])
        self.assertFalse(self.settings["tracklist.play-count.visible"])
        self.assertTrue(self.settings["tracklist.last-started.visible"])
        self.assertFalse(self.settings["tracklist.last-played.visible"])
        self.assertTrue(self.settings["tracklist.filepath.visible"])
